header {
/*
 * Copyright (c) 2004, 2005, 2006 The University of Utah
 *
 * This file is part of Jiazzi.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

  package edu.utah.jiazzi.parser;
   
  import java.util.*;

  import edu.utah.jiazzi.loc.LocPackage;
  import edu.utah.jiazzi.loc.LocRoot;
  import edu.utah.jiazzi.loc.Local;
  import edu.utah.jiazzi.unit.ArgVar;
  import edu.utah.jiazzi.unit.AtomUnit;
  import edu.utah.jiazzi.unit.CompoundUnit;
  import edu.utah.jiazzi.unit.Unit;
  import edu.utah.jiazzi.unit.UnitClass;
  import edu.utah.jiazzi.unit.UnitEnvironment;
  import edu.utah.jiazzi.unit.UnitOpen;
  import edu.utah.jiazzi.unit.UnitPackage;
  import edu.utah.jiazzi.unit.UnitRoot;
  import edu.utah.jiazzi.unit.UnitSignature;
  import edu.utah.jiggy.clazz.PackageModel;
  import edu.utah.jiggy.meta.flag.FlagConstants;
  import edu.utah.jiggy.meta.name.ClassName;
  import edu.utah.jiggy.meta.name.ExpandedName;
  import edu.utah.jiggy.meta.name.MemberName;
  import edu.utah.jiggy.meta.name.Name;
  import edu.utah.jiggy.meta.name.PackageName;
  import edu.utah.jiggy.meta.name.VariableName;
  import edu.utah.jiggy.meta.shape.ArgShape;
  import edu.utah.jiggy.meta.shape.Builder;
  import edu.utah.jiggy.meta.shape.ClassShape;
  import edu.utah.jiggy.meta.shape.FieldShape;
  import edu.utah.jiggy.meta.shape.InitShape;
  import edu.utah.jiggy.meta.shape.MethodShape;
  import edu.utah.jiggy.meta.shape.NameTypeShape;
  import edu.utah.jiggy.meta.shape.ProcedureShape;
  import edu.utah.jiggy.meta.shape.ThrowzShape;
  import edu.utah.jiggy.meta.type.ClassType;
  import edu.utah.jiggy.meta.type.PackageType;
  import edu.utah.jiggy.meta.type.PrimitiveType;
  import edu.utah.jiggy.meta.type.RefType;
  import edu.utah.jiggy.meta.type.Root;
  import edu.utah.jiggy.meta.type.TypeConstants;
  import edu.utah.jiggy.meta.type.TypeType;
  import edu.utah.jiggy.meta.type.VariableType;
}
/*
 * Make sure to run antlr.Tool on the lexer.g file first!
 */
options {
  mangleLiteralPrefix = "JIAZZI_";
}

class JParser extends Parser;
options {
  importVocab=JIAZZI; // use vocab generated by lexer
  k=6;
} {
  private UnitEnvironment env;
  public void setEnvironment(UnitEnvironment env) {
    this.env = env;
  }
  public void reportError(String str) {
    this.env.err().println("" + str);
    this.env.fail();
  }
  public void reportError(antlr.RecognitionException exc) {
    this.reportError(exc.toString());
  }
}


keyword_r[] returns [String id] {
  id = null;
}: (JIAZZI_empty { id = "empty"; } 
    |
    JIAZZI_name { id = "name"; } 
    |
    JIAZZI_arg { id = "arg"; }   
    |
    JIAZZI_open { id = "open"; } 
    |
    JIAZZI_bind { id = "bind"; } 
    |
    JIAZZI_link { id = "link"; } 
    |
    JIAZZI_local { id = "local"; } 
    |
    JIAZZI_to { id = "to"; } 
    | 
    JIAZZI_unit { id = "unit"; } 
    |
    JIAZZI_atom { id = "atom"; } 
    |
    JIAZZI_compound { id = "compound"; } 
    |
    JIAZZI_signature { id = "signature"; } 
    | 
    JIAZZI_generic { id = "generic"; } 
    |
    JIAZZI_export { id = "export"; } 
    |
    JIAZZI_external { id = "external"; }
    );



var_r[] returns [String str] {
  str = null;
}: id:ID {
  str = id.getText();
  if (str.charAt(0) != '[' ||
      str.charAt(str.length() - 1) != ']')
    throw new antlr.RecognitionException("expecting single bracked id, not \"" + id + "\"");
  str = str.substring(1, str.length() -1);
  if (str.indexOf('[') != -1 ||
      str.indexOf(']') != -1)
    throw new antlr.RecognitionException("expecting single bracked id, not \"" + id + "\"");
};

expanded0_r[UnitEnvironment env, UnitRoot root] returns [ExpandedName exp] {
  exp = null;
  String text;
}: ((id:ID {
  text = id.getText();
  if (text.indexOf('[') == -1) { // ']'
    exp = new ExpandedName(text, VariableName.NULL, "");
  } else {
    String btext = text.substring(0, text.indexOf('['));
    String etext = text.substring(text.indexOf(']') + 1, text.length());
    String vtext = text.substring(text.indexOf('[') + 1, text.indexOf(']'));
    if (root instanceof UnitSignature && 
	((UnitSignature) root).lexemes().containsKey(vtext)) {
      UnitSignature sig = (UnitSignature) root;
      exp = new ExpandedName("", sig.findNameVar(vtext), "");
      exp = exp.prepend(btext).append(etext);
    } else {
      this.reportError("cannot find name variable " + vtext + " in " + root);
    }
  }
}) | ({
} text=keyword_r[] {
  exp = new ExpandedName(text, VariableName.NULL, "");
}));

expanded_r[UnitEnvironment env, UnitRoot root] returns [ExpandedName exp] {
  exp = null;
}: (exp=expanded0_r[env,root]) | (DQUOTE DQUOTE {
  exp = new ExpandedName("", VariableName.NULL, "");
});

name_r[UnitEnvironment env, UnitRoot root, Name nm0] returns [Name nm] {
  nm = nm0;
  ExpandedName exp; 
}: exp=expanded0_r[env,root] {
  nm = nm.setExpanded(exp);
};

longid_r[] returns [String id] {
  id = null;
  String id0;
}: id=id_r[] (DOT id0=id_r[] {
  id = id + '.' + id0;
})*;

id_r[] returns [String id] {
  id = null;
}: ((id0:ID { 
  id = id0.getText(); 
  if (id.indexOf('[') != -1) 
    throw new antlr.RecognitionException("not expecting variable \"" + id + "\"");
}) | id=keyword_r[]);


classname_r[UnitEnvironment env, UnitRoot unit] returns [ClassName cname] {
  cname = new ClassName();
  Name nm;
  ClassName CONSTANT = new ClassName("");
}: nm=name_r[env,unit,CONSTANT] {
  cname = (ClassName) nm;
} ((DOLLAR) {
} nm=name_r[env,unit,CONSTANT] {
  cname = ((ClassName) nm).setOuter(cname);
})*;

pkgname_r[] returns [PackageName pname] {
  pname = null;
  String pn;
}: pn=longid_r[] {
  pname = new PackageName(pn);
};

pkg_decl_r[UnitEnvironment env, UnitRoot root] returns [UnitPackage upkg] {
  upkg = null;
  PackageName pkgname;
  String signame;
}: pkgname=pkgname_r[] {
  if (root.packages().containsKey(pkgname)) {
    upkg = (UnitPackage) root.packages().get(pkgname);
    this.reportError(upkg + " already package in " + root);
     
  } else upkg = new UnitPackage(root, pkgname);
} (JIAZZI_extends pkgname=pkgname_r[] {
  if (!root.packages().containsKey(pkgname)) {
    this.reportError(pkgname + " not found in " + root);
     
  } else {
    UnitPackage epkg = (UnitPackage) root.packages().get(pkgname);
    upkg.setExtended(epkg);
  }
})? (COLON signame=longid_r[] {
  upkg.setSignature(env.findSignature(signame));
})?;

lexeme_decl_r[UnitEnvironment env, UnitSignature root] returns [VariableName var] {
  var = null;
  String name;
}: name=var_r[] {
  var = root.findNameVar(name);
  if (var != null) {
    this.reportError(var + " already lexeme in " + root);
     
  } else {
    var = root.newLexeme(name);
  }
};

arg_decl_r[UnitEnvironment env, UnitSignature root] returns [ArgVar var] {
  var = null;
  String name;
}: name= var_r[] {
  var = root.findArgVar(name);
  if (var != null) {
    var = (ArgVar) root.argVars().get(name);
    this.reportError(var + " already argument paramater in " + root);
     
  } else {
    var = root.newArgVar(name);
  }
};


generic_decl_r[UnitEnvironment env, UnitRoot root] returns [VariableType var] {
  var = null;
  String name;
  RefType bound=null;
}: name=id_r[] ({
  bound = env.platform().findClass(new ClassName("java.lang.Object"));
} | (COLON bound=generic_bind_type_r[env,root,root])) {
  if (root.generics().containsKey(name)) {
    var = (VariableType) root.generics().get(name);
    this.reportError(var + " already generic in " + root);
     
  } else if (!(bound instanceof ClassType)) {
    this.reportError("bound for " + var + " must be class type");
     
  } else {
    var =  root.newGeneric(name, (ClassType) bound);
  }
};


bind_upkg_r[UnitEnvironment env, UnitRoot root] returns [UnitPackage upkg] {
  PackageName pkgname;
  upkg = null;
}: pkgname=pkgname_r[] {
  if (!root.packages().containsKey(pkgname)) {
    this.reportError("cannot find package " + pkgname + " in " + root);
     
  } else {
    upkg = (UnitPackage) root.packages().get(pkgname);
    if (upkg.sig() == null) {
      this.reportError(upkg + " in " + root + " does not have signature so cannot be bound");
       
    }
  }
};

bind_pkg_r[UnitEnvironment env, UnitRoot root] {
  PackageType fpkg=null;
  UnitPackage upkg=null;
  UnitPackage tpkg=null;
  PackageName pkgname=null;
}: ((WILD JIAZZI_to upkg=bind_upkg_r[env,root] AT WILD {
  for (Iterator i = upkg.sig().params().iterator(); i.hasNext(); /* i = i */) {
    tpkg = (UnitPackage) i.next();
    if (upkg.pkgBindings().containsKey(tpkg)) continue;
    if (!root.packages().containsKey(tpkg.name())) continue;
    fpkg = (PackageType) root.packages().get(tpkg.name());
    if (fpkg == null) continue;
    if (!((UnitPackage) fpkg).isParam()) continue;
    upkg.bind(tpkg, fpkg);
  }
}) | (pkgname=pkgname_r[] JIAZZI_to {
  if (root.packages().containsKey(pkgname) &&
      ((UnitPackage) root.packages().get(pkgname)).isParam()) {
	fpkg = (UnitPackage) root.packages().get(pkgname);
	if (fpkg == null) throw new Error("what the hell");
      } else {
	if (!env.platform().containsPackage(pkgname)) {
	  this.reportError("cannot find package " + pkgname + " in " + root + " or " + env.platform());
	  fpkg = null;
	} else {
	  fpkg = env.platform().findPackage(pkgname);
	}
      }
} ((upkg=bind_upkg_r[env,root] AT pkgname=pkgname_r[] {
  if (upkg.isOpen()) {
    boolean pass = false;
    UnitPackage ops[] = { upkg.open().in(), upkg.open().out(), upkg.open().fixed() };
    for (int i = 0; i < ops.length; i++) {
      if (!ops[i].sig().packages().containsKey(pkgname)) continue;
      tpkg = (UnitPackage) ops[i].sig().packages().get(pkgname);
      if (!tpkg.isParam()) continue;
      if (fpkg != null) 
	ops[i].bind(tpkg, fpkg);
      pass = true;
    }
    if (!pass) {
      this.reportError(pkgname + " is not a package parameter of " + upkg.sig() + " for " + upkg);
       
    }
  } else {
    if (!upkg.sig().packages().containsKey(pkgname)) {
      this.reportError(pkgname + " is not a package parameter of " + upkg.sig() + " for " + upkg);
       
      fpkg = null;
    } 
    tpkg = (UnitPackage) upkg.sig().packages().get(pkgname);
    if (tpkg != null && !tpkg.isParam()) {
      this.reportError(tpkg + " is not a package parameter");
       
    } else {
      if (tpkg != null && !upkg.pkgBindings().containsKey(tpkg) && fpkg != null) 
	upkg.bind(tpkg, fpkg);
    }
  }
}) | (WILD AT pkgname=pkgname_r[] {
  for (Iterator i = root.packages().values().iterator(); i.hasNext(); /* i = i */) {
    upkg = (UnitPackage) i.next();
    if (upkg.sig() == null) continue;
    if (!upkg.sig().packages().containsKey(pkgname)) continue;
    tpkg = (UnitPackage) upkg.sig().packages().get(pkgname);
    if (!tpkg.isParam()) continue;
    if (upkg.pkgBindings().containsKey(tpkg)) continue;
    if (fpkg != null) 
      upkg.bind(tpkg, fpkg);
  }
}))));

/* bind name parameters and arg parameters
 */
sigbind_r[UnitEnvironment env, UnitSignature usig] {
  ExpandedName exp;
  ProcedureShape proc;
  String vname;
}: /*AT vname=id_r[]*/ vname=var_r[] EQUAL ((exp=expanded_r[env,usig] {
  if (!usig.bindNameVar(vname, exp)) {
    this.reportError(usig + " does not have a argument parameter " + vname);
     
  }
}) | (proc=proc_r[env,usig] {
  if (!usig.bindArgVar(vname, proc)) {
    this.reportError(usig + " does not have a argument parameter " + vname);
     
  }
}));

/*
bind_lexeme_r[UnitEnvironment env, UnitRoot root] {
  ExpandedName exp;
  UnitPackage upkg;
  VariableName var;
  String vname;
}: ((WILD JIAZZI_to upkg=bind_upkg_r[env,root] AT WILD {
  for (Iterator i = root.lexemes().values().iterator(); i.hasNext(); i = i) {
    VariableName fvar = (VariableName) i.next();
    if (!upkg.sig().lexemes().containsKey(fvar.text())) continue;
    var = upkg.sig().lexemes().get(fvar.text());
    if (upkg.lexemeBindings().containsKey(var)) continue;
    upkg.bind(var, new ExpandedName("", fvar, ""));
  }
}) | (exp=expanded_r[env,root] JIAZZI_to {
} ((upkg=bind_upkg_r[env,root] AT vname=id_r[] {
  if (!upkg.sig().lexemes().containsKey(vname)) {
    this.reportError(upkg + " does not have a lexeme parameter " + vname);
     
  } else {
    var = upkg.sig().lexemes().get(vname);
    if (!upkg.lexemeBindings().containsKey(var))
      upkg.bind(var, exp);
  }
}) | (WILD AT vname=id_r[] {
  for (Iterator i = root.packages().values().iterator(); i.hasNext(); i = i) {
    upkg = (UnitPackage) i.next();
    if (upkg.sig() == null) continue;
    if (!upkg.sig().lexemes().containsKey(vname)) continue;
    var = upkg.sig().lexemes().get(vname);
    if (!upkg.lexemeBindings().containsKey(var)) 
      upkg.bind(var, exp);
  }
}))));
*/

bind_generic_r[UnitEnvironment env, UnitRoot root] {
  RefType ref;
  UnitPackage upkg;
  VariableType var;
  String vname;
}: ((WILD JIAZZI_to upkg=bind_upkg_r[env,root] AT WILD {
  for (Iterator i = root.generics().values().iterator(); i.hasNext(); /* i = i */) {
    VariableType fvar = (VariableType) i.next();
    if (!upkg.sig().generics().containsKey(fvar.name())) continue;
    var = (VariableType) upkg.sig().generics().get(fvar.name());
    if (upkg.genericBindings().containsKey(var)) continue;
    upkg.bind(var, fvar);
  }
}) | (ref=generic_bind_type_r[env,root,root] JIAZZI_to {
} ((upkg=bind_upkg_r[env,root] AT vname=id_r[] {
  if (!upkg.sig().generics().containsKey(vname)) {
    this.reportError(upkg + " does not have a generic parameter " + vname);
     
  } else {
    var = (VariableType) upkg.sig().generics().get(vname);
    if (!upkg.genericBindings().containsKey(var))
      upkg.bind(var, ref);
  }
}) | (WILD AT vname=id_r[] {
  for (Iterator i = root.packages().values().iterator(); i.hasNext(); /* i = i */) {
    upkg = (UnitPackage) i.next();
    if (upkg.sig() == null) continue;
    if (!upkg.sig().generics().containsKey(vname)) continue;
    var = (VariableType) upkg.sig().generics().get(vname);
    if (!upkg.genericBindings().containsKey(var)) 
      upkg.bind(var, ref);
  }
}))));

signature_r[UnitEnvironment env] returns [UnitSignature sig] {
  sig = null;
  String name, pname = null;
  UnitPackage upkg;
  VariableType generic;
  VariableName lexeme;
  ArgVar arg;
  int n = 0;
  PackageName pkgname;
  // List<PackageModel> old = new ArrayList<PackageModel>(env.defaults);
  List old = new ArrayList(env.defaults);
  env.defaults.clear();
  // if (!env.defaults.isEmpty()) throw new Error("unexpected");
  env.defaults.add((PackageModel) env.platform().findPackage(new PackageName("java.lang")));
}: (JIAZZI_external (JIAZZI_package)? pkgname=pkgname_r[] {
  env.defaults.add((PackageModel) env.platform().findPackage(pkgname));
} SEMI)* JIAZZI_signature name=longid_r[] {
  sig = new UnitSignature(name);
} EQUAL (((pname=id_r[] COLON) | ({ pname = "$$" + n; n++; })) name=longid_r[] {
  UnitSignature zsig = env.findSignature(name);
  UnitPackage zpkg = new UnitPackage(sig, new PackageName(pname));
  zpkg.setSignature(zsig);
  sig.makeSuper(zpkg);
} PLUS)* (LCURLY
  (((JIAZZI_package upkg=pkg_decl_r[env,sig] { sig.makeParam(upkg); }
     (COMMA upkg=pkg_decl_r[env,sig] { sig.makeParam(upkg); })*) |
    ((JIAZZI_name) lexeme=lexeme_decl_r[env,sig]
     (COMMA lexeme=lexeme_decl_r[env,sig])*) |
    ((JIAZZI_arg) arg=arg_decl_r[env,sig]
     (COMMA arg=arg_decl_r[env,sig])*) |
    (JIAZZI_generic generic=generic_decl_r[env,sig]
     (COMMA generic=generic_decl_r[env,sig])*)) SEMI)*
  (sigbind_r[env,sig] SEMI)*
  (JIAZZI_bind
   ((JIAZZI_package ((COMMA)? bind_pkg_r[env,sig])*) |
    /* (JIAZZI_name ((COMMA)? bind_lexeme_r[env,sig])*) */

    /* (JIAZZI_lexeme ((COMMA)? bind_lexeme_r[env,sig])*) | */
    (JIAZZI_generic ((COMMA)? bind_generic_r[env,sig])*)) SEMI)*
  (outerclass_r[env, sig])* RCURLY) {
  env.defaults.clear();
  env.defaults.addAll(old);
}; exception catch [antlr.TokenStreamException e] {
    this.reportError("" + e);
     
  }

outerclass_r[UnitEnvironment env, UnitSignature sig] {
  Builder bld = new Builder();
  UnitClass ucls;
}: flags_r[bld] ucls=class_r[env, sig, bld, null];

membername_r[UnitEnvironment env, UnitSignature sig] returns [MemberName nm] {
  nm = new MemberName();
  Name nm0;
}: nm0=name_r[env, sig, nm] {
  nm = (MemberName) nm0;
};

flags_r[Builder shape] {
  shape.set(FlagConstants.PUBLIC);
  shape.set(FlagConstants.CONCRETE);
  /* bld.set(FlagConstants.INSTANCE); */
}: (JIAZZI_public | JIAZZI_protected { shape.set(FlagConstants.PROTECTED); } |
    JIAZZI_abstract { shape.set(FlagConstants.ABSTRACT); } |
    JIAZZI_final { shape.set(FlagConstants.FINAL); } |
    JIAZZI_static { shape.set(FlagConstants.STATIC); })*;

class_r[UnitEnvironment env, UnitSignature sig, Builder bld, UnitClass outer] 
  returns [UnitClass ucls] {
  ClassName cname=new ClassName("");
  ucls = null;
  ClassShape shape=null;
  ClassType zuper=null;
  Builder bld1 = new Builder();
  UnitClass inner=null;
  Name nm = null;
}: (JIAZZI_interface {
  bld.set(FlagConstants.INTERFACE);
  bld.set(FlagConstants.ABSTRACT);
} | JIAZZI_class {
}) nm=name_r[env,sig,cname] {
  cname = (ClassName) nm;
  if (outer != null) cname = cname.setOuter(outer.name());
  if (sig.self().containsClass(cname.nopkg())) {
    ucls = (UnitClass) sig.self().classes().get(cname.nopkg());
  } else {
    ucls = new UnitClass(sig.self(), cname.nopkg());
  }
  if (ucls.isShapeSet()) {
    shape = ucls.shape();
  } else {
    shape = new ClassShape();
    ucls.setShape(shape);
    shape.add(bld);
  }
} (JIAZZI_extends ((COMMA)? zuper=supertype_r[env, sig] {
  if (shape.binaries().contains(FlagConstants.INTERFACE))
    shape.addInterface(zuper);
  else shape.setZuper(zuper);
})*)? (JIAZZI_implements ((COMMA)? zuper=supertype_r[env, sig] {
  shape.addInterface(zuper);
})*)? (PLUS EQUAL)? (LCURLY ({
  bld1 = new Builder();
} flags_r[bld1] {
} ((inner=class_r[env, sig, bld1, ucls] {
  
}) | (member_r[env,sig,shape,bld1]) SEMI))* RCURLY);



constant_r[] returns [Object cnst] {
  cnst = null;
  boolean minus = false;
}: EQUAL (((MINUS { minus = true; })? 
  (a:INTEGER_TEN {
    cnst = Integer.valueOf(a.getText(), 10);
  } /* | b:INTEGER_OCT {
       cnst = Integer.valueOf(b.getText().substring(1, b.getText().length()), 8); 
       } */ 
  | c : INTEGER_HEX {
    cnst = Integer.valueOf(c.getText().substring(2, c.getText().length()), 16);
  }) {
  if (minus) cnst = new Integer(-((Integer) cnst).intValue());
}) | (JIAZZI_true { cnst = new Boolean(true); } | JIAZZI_false { cnst = new Boolean(false); }));

member_r[UnitEnvironment env, UnitSignature sig, ClassShape shape, Builder bld] {
  ExpandedName exp;
  NameTypeShape nt = null;
  ProcedureShape proc = null;
  ThrowzShape throwz = null;
  Object cnst = null;
}: (
    // Member methods and fields.
    nt=nt_r[env,sig]
    (proc=proc_r[env,sig] throwz=throwz_r[env,sig]
    | (BRACKETS { if (nt != null) nt = nt.setType(nt.type().up()); })*
      (cnst=constant_r[])?
    )
   |
    // Constructors.
    // XXX: the value of `exp' is never used.  It should be checked.
    exp=expanded0_r[env,sig] proc=proc_r[env,sig] throwz=throwz_r[env,sig]
   |
    // "Unnamed" constructors.
    // XXX: is this syntax really needed?  It seems gratuitous.
    proc=proc_r[env,sig] throwz=throwz_r[env,sig]
   ) {
  if (false) {
  } else if (nt == null && proc != null) {
    InitShape init = new InitShape();
    init.add(bld);
    init.set(proc); init.set(throwz);
    shape.add(init);
  } else if (proc == null && nt != null) {
    FieldShape fld = new FieldShape();
    fld.add(bld);
    fld.set(nt);
    if (shape.binaries().contains(FlagConstants.INTERFACE)) {
      fld.set(FlagConstants.STATIC);
      fld.set(FlagConstants.IMMUTABLE);
    }
    if (cnst != null) {
      if (cnst instanceof Integer && !nt.type().equals(TypeConstants.INT)) {
	 
	this.reportError("integer compile time constant cannot be added to " + fld);
      }
      if (cnst instanceof Boolean && !nt.type().equals(TypeConstants.BOOLEAN)) {
	 
	this.reportError("boolean compile time constant cannot be added to " + fld);
      }
      if (!fld.isSet(FlagConstants.STATIC) || !fld.isSet(FlagConstants.IMMUTABLE)) {
	 
	this.reportError("compile time constant cannot be added to " + fld);
      }
      fld.setConstant(cnst);
    }
    shape.add(fld);
  } else if (proc != null && nt != null) {
    MethodShape mthd = new MethodShape();
    mthd.add(bld);
    mthd.set(nt); mthd.set(proc); mthd.set(throwz);
    if (shape.binaries().contains(FlagConstants.INTERFACE)) {
      mthd.set(FlagConstants.ABSTRACT);
    }
    shape.add(mthd);
  }
};



nt_r[UnitEnvironment env, UnitSignature sig] returns [NameTypeShape nt] {
  nt = new NameTypeShape();
  TypeType tp;
  MemberName mmbr;
}: tp=type_r[env,sig] mmbr=membername_r[env,sig] {
  nt = nt.setType(tp).setName(mmbr);
};

proc_r[UnitEnvironment env, UnitSignature sig] returns [ProcedureShape proc] {
  proc = new ProcedureShape();
  String vname;
}: LPAREN (proc=arg_r[env,sig,proc] ((COMMA) proc=arg_r[env,sig,proc])*)? RPAREN (vname=var_r[] {
  ArgVar var = sig.findArgVar(vname);
  if (var == null) {
     
    this.reportError("cannot find arg parameter \"" + vname + "\" in " + sig);
  } else {
    proc = proc.set(var);
  }
})?;

throwz_r[UnitEnvironment env, UnitSignature sig] returns [ThrowzShape throwz] {
  ClassType ctype;
  throwz = new ThrowzShape();
}: (JIAZZI_throws ctype=supertype_r[env,sig] {
  throwz = throwz.add(ctype);
} ((COMMA) ctype=supertype_r[env,sig] {
  throwz = throwz.add(ctype);
})*)?;

arg_r[UnitEnvironment env, UnitSignature sig, ProcedureShape proc0] returns [ProcedureShape proc] {
  proc = proc0;
  TypeType tp;
  String name;
}: tp=type_r[env,sig] (name=id_r[] (BRACKETS {
  tp = tp.up();
})* {
  proc = proc.addArg((new ArgShape()).setType(tp).setName(name));
} | {
  proc = proc.addArg(tp);
});

type_r[UnitEnvironment env, UnitSignature sig] returns [TypeType tp] {
  tp = null;
}: (tp=primitive_r[] | tp=reftype0_r[env,sig]) (BRACKETS {
  tp = tp.up();
})*;

reftype_r[UnitEnvironment env, UnitRoot root] returns [RefType tref] {
  tref = null;
  PrimitiveType tp;
}: ((tp=primitive_r[] BRACKETS {
  tref = tp.up();
}) | tref=reftype0_r[env,root]) (BRACKETS {
  tref = tref.up();
})*;

supertype_r[UnitEnvironment env, UnitSignature sig] returns [ClassType ctype] {
  String name;
  ctype = null;
  ClassName cname = null;
  ClassName inner;
}: name=id_r[] (ctype=classtype_r[env, sig, name] | ({
  // PackageType pkg = env.platform().findPackage(new PackageName("java.lang"));
  cname = new ClassName(name);
} (DOLLAR inner=classname_r[env,sig] {
  cname = inner.setOuter(cname);
})? {
  if (sig.self().containsClass(cname)) {
    ctype = sig.self().findClass(cname);
  } else  for (Iterator i = env.defaults.iterator(); i.hasNext(); /* i = i */) {
    PackageModel pkg = (PackageModel) i.next();
    if (pkg.containsClass(cname)) ctype = pkg.findClass(cname);
    if (ctype != null) break;
  }
  if (ctype == null) {
    if (sig.self().containsClass(cname)) {
      ctype = sig.self().findClass(cname);
    } else {
      // UnitClass ucls =
      new UnitClass(sig.self(), cname);
      ctype = sig.self().findClass(cname);
    }
  }
  // this.reportError("cannot find class " + cname + " in java.lang");
  //  
  // ctype = pkg.findClass(new ClassName("Object"));
}));

reftype0_r[UnitEnvironment env, UnitRoot root] returns [RefType ref] {
  ref = null;
  ExpandedName exp;
  String name = null;
  RefType param;
  ClassName inner;
  ClassName cname = null;
}: exp=expanded0_r[env,root] (({
  if (exp.var() != VariableName.NULL) {
      
    this.reportError("not expecting variable in package name " + exp);
  }
  name = exp.extract();
} ref=classtype_r[env, root, name]) | ({
 cname = (ClassName) (new ClassName()).setExpanded(exp);
} ((DOLLAR) inner=classname_r[env,root] {
  cname = inner.setOuter(cname);
})? {
  name = null;
  if (exp.var() == VariableName.NULL) name = exp.extract();
  // PackageType jlpkg = env.platform().findPackage(new PackageName("java.lang"));
  // System.err.println("XXX " + name + " " + exp + " " + exp.var() + " " + root.generics().keySet());

  ref = null;
  if (name != null && root.generics().containsKey(name)) {
    ref = (VariableType) root.generics().get(name);
  } else if (root instanceof UnitSignature &&
	     ((UnitSignature) root).self().containsClass(cname)) {
    ref = ((UnitSignature) root).self().findClass(cname);
  } else for (Iterator i = env.defaults.iterator(); i.hasNext(); /* i = i */) {
    PackageModel pkg = (PackageModel) i.next();
    if (pkg.containsClass(cname)) {
      ref = pkg.findClass(cname);
      break;
    }
  }
  if (ref == null && (root instanceof UnitSignature)) {
    UnitSignature usig = (UnitSignature) root;
    if (usig.self().containsClass(cname)) {
      ref = usig.self().findClass(cname);
    } else { 
      // UnitClass ucls =
      new UnitClass(usig.self(), cname);
      ref = usig.self().findClass(cname);
    }
  } 
  if (ref == null) {
    this.reportError("cannot find class " + cname + " in " + root + "'s generics or in java.lang");
    ref = env.platform().findPackage(new PackageName("java.lang")).findClass(new ClassName("Object"));
  }
} ((LT ((COMMA)? param=reftype_r[env,root] {
  ref = ((ClassType) ref).addParam(param);
})+ GT)?)));

classtype_r[UnitEnvironment env, UnitRoot root, String name] returns [ClassType ctype] {
  ctype = null;
  String sname;
  PackageType pkg = null;
}: ((DOT sname=id_r[] {
  name = name + '.' + sname;
})* {
  PackageName pname = new PackageName(name);
  if (root.packages().containsKey(pname) &&
      ((UnitPackage) root.packages().get(pname)).isParam()) pkg = (UnitPackage) root.packages().get(pname);
  else if (env.platform().containsPackage(pname)) pkg = env.platform().findPackage(pname);
  else {
    this.reportError("cannot find package " + pname + " in " + root + " or classpath " + env.platform());
    pkg = null;
  }
} DOT) ctype=classtype1_r[env, root, pkg];

classtype1_r[UnitEnvironment env, UnitRoot root, PackageType pkg] returns [ClassType ctype] {
  ClassName cname;
  ctype=null;
  RefType param;
}: cname=classname_r[env,root] {
  if (pkg != null && pkg.containsClass(cname)) {
    ctype = pkg.findClass(cname);
  } else {
    if (pkg != null && !pkg.containsClass(cname)) {
      this.reportError("cannot find " + cname + " in " + pkg);
    }
    ctype = env.platform().findClass(new ClassName("java.lang.Object"));
  }
} (LT ((COMMA)? param=reftype_r[env,root] {
  ctype = ((ClassType) ctype).addParam(param);
})+ GT)?;

primitive_r[] returns [PrimitiveType tp] {
  tp = null;
}: (JIAZZI_boolean { tp = TypeConstants.BOOLEAN; } |
    JIAZZI_byte    { tp = TypeConstants.BYTE;    } |
    JIAZZI_char    { tp = TypeConstants.CHAR;    } |
    JIAZZI_short   { tp = TypeConstants.SHORT;   } |
    JIAZZI_int     { tp = TypeConstants.INT;     } |
    JIAZZI_long    { tp = TypeConstants.LONG;    } |
    JIAZZI_float   { tp = TypeConstants.FLOAT;   } |
    JIAZZI_double  { tp = TypeConstants.DOUBLE;  } |
    JIAZZI_void    { tp = TypeConstants.VOID;    });

unit_r[UnitEnvironment env] returns [Unit unit] {
  unit = null;
}: (JIAZZI_atom unit=atom_r[env] | 
    JIAZZI_compound unit=compound_r[env]); exception catch [antlr.TokenStreamException e] {
      this.reportError("" + e);
       
    }


atom_r[UnitEnvironment env] returns [AtomUnit atom] {
  atom = null;
  String uname; 
}: (uname=longid_r[] {
  atom = new AtomUnit(uname);
} (LCURLY unitsignature_r[env,atom]) {} RCURLY);

compound_r[UnitEnvironment env] returns [CompoundUnit compound] {
  compound = null;
  String uname;   
}: (uname=longid_r[] {
  compound = new CompoundUnit(uname);
} (LCURLY unitsignature_r[env,compound] {
} (RCURLY LCURLY)? linkage_r[env,compound]) RCURLY);

unitsignature_r[UnitEnvironment env, Unit unit] {
  VariableType generic=null;
  // VariableName lexeme=null;
  UnitPackage upkg=null;
  int mode = -1;
  final int IMPORT = 0;
  final int EXPORT = 1;
}: (
    ((
      (JIAZZI_generic generic=generic_decl_r[env,unit] (COMMA generic=generic_decl_r[env,unit])*) |
      (open_r[env,unit]) |
      (((JIAZZI_import { mode = IMPORT; }) | (JIAZZI_export { mode = EXPORT; }))
       upkg=pkg_decl_r[env,unit] {
	switch (mode) {
	case IMPORT: unit.makeImport(upkg); break;
	case EXPORT: unit.makeExport(upkg); break;
	default: throw new Error("stupid");
	}
      })) SEMI)*
     (JIAZZI_bind
      ((JIAZZI_package ((COMMA)? bind_pkg_r[env,unit])*) |
       /* ((JIAZZI_lexeme | JIAZZI_name) ((COMMA)? bind_lexeme_r[env,unit])*) | */
       (JIAZZI_generic ((COMMA)? bind_generic_r[env,unit])*)) SEMI)*
    );

open_r[UnitEnvironment env, Unit unit] {
  UnitOpen open = null;
  String pname = null;
  String sid = null;
}: (JIAZZI_open pname=id_r[] COLON {
    open = new UnitOpen(unit, pname);
  } ((AT JIAZZI_empty) | (sid=longid_r[] {
    UnitSignature usig = env.findSignature(sid);
    open.in().setSignature(usig);
  })) MINUS GT sid=longid_r[] {
    UnitSignature usig = env.findSignature(sid);
    open.out().setSignature(usig);
  });


generic_bind_type_r[UnitEnvironment env, UnitRoot root, Root search] returns [RefType ctype] {
  ctype = null;
  ClassName cname=null;
  PackageType pkg=null;
  PackageName pkgname = null;
  String part=null;
}: part=id_r[] ((DOT {
  pkgname = new PackageName(part);
} (part=id_r[] DOT {
  pkgname = (new PackageName(part)).setOuter(pkgname);
})* {
  if (search.containsPackage(pkgname)) {
    pkg = search.findPackage(pkgname);
    if (search instanceof UnitRoot) {
      if (!((UnitPackage) pkg).isParam()) {
	 
	this.reportError(pkg + " is not a parameter package");
      }
      ((UnitPackage) pkg).resolve(env, UnitPackage.CLASS_RESOLVED);
    } else {
      if (!((LocPackage) pkg).isSource()) {
	 
	this.reportError(pkg + " is not a source package");
      }
      /* resolve the package...hope its variables are bound appropriately. */
      ((LocPackage) pkg).resolveInit(env);
    }
  } else {
    pkg = env.platform().findPackage(pkgname);
    if (pkg == null) {
       
      this.reportError("cannot find " + pkg + " in " + root + " or " + env.platform());
      pkg = env.platform().findPackage(new PackageName("java.lang"));
    }
  }
} ctype=classtype1_r[env,root,pkg]) | ({
  pkg = env.platform().findPackage(new PackageName("java.lang"));
  cname = new ClassName(part);
  if (pkg.containsClass(cname)) {
    ctype = pkg.findClass(cname);
  } else if (root.generics().containsKey(part)) {
    ctype = (RefType) root.generics().get(part);
  } else {
     
    this.reportError("cannot find " + part + " in java.lang or generics");
    ctype = pkg.findClass(new ClassName("Object"));
  }
}));

local_decl_r[UnitEnvironment env, CompoundUnit unit] {
  String lname = null;
  String uname;
}: (lname=id_r[] COLON | { lname = null; }) uname=longid_r[] {
  Unit lunit = env.findUnit(uname);
  Local local;
  if (lname == null) local = unit.newLocal(lunit);
  else local = unit.newLocal(lname, lunit);
  local.resolveInit(env);
}; 

linkage_r[UnitEnvironment env, CompoundUnit unit] {
  unit.self().resolveInit(env);
}:  (((  (/*JIAZZI_local |*/ (JIAZZI_link JIAZZI_unit)) local_decl_r[env, unit] ((COMMA) local_decl_r[env, unit])*) |
         (JIAZZI_link JIAZZI_package link_pkg_r[env, unit] ((COMMA) link_pkg_r[env, unit])*) |
         (JIAZZI_link JIAZZI_generic link_generic_r[env, unit] ((COMMA) link_generic_r[env, unit])*)) 

     SEMI)*;



local_r[UnitEnvironment env, CompoundUnit unit] returns [Local local] {
  local = null;
  String lname;
}: lname=id_r[] {
  if (!unit.locals().containsKey(lname)) {
     
    this.reportError("cannot find local " + lname + " in " + unit);
  } else local = (Local) unit.locals().get(lname);
};

link_generic_r[UnitEnvironment env, CompoundUnit unit] {
  RefType ref;
  Local local;
  LocRoot source = null;
  String vname;
}: ((WILD JIAZZI_to local=local_r[env,unit] AT WILD {
  if (local != null) {
    for (Iterator i = local.unit().generics().values().iterator(); i.hasNext(); /* i = i */) {
      VariableType var = (VariableType) i.next();
      if (!unit.generics().containsKey(var.name())) continue;
      if (local.genericLinkage().containsKey(var)) continue;
      local.link(var, (VariableType) unit.generics().get(var.name()));
    }
  }
}) | (((source=local_r[env,unit] AT) | { source = unit.self(); })
      ref=generic_bind_type_r[env,unit,source] JIAZZI_to ((WILD AT vname=id_r[] {
  for (Iterator i = unit.locals().values().iterator(); i.hasNext(); /* i = i */) {
    local = (Local) i.next();
    if (!local.unit().generics().containsKey(vname)) continue;
    VariableType var = (VariableType) local.unit().generics().get(vname);
    if (local.genericLinkage().containsKey(var)) continue;
    local.link(var, ref);
  }
}) | (local=local_r[env,unit] AT vname=id_r[] {
  if (local.unit().generics().containsKey(vname)) {
    VariableType var = (VariableType) local.unit().generics().get(vname);
    if (!local.genericLinkage().containsKey(var)) {
      local.link(var, ref);
    }
  }
}))));


/* 
link_lexeme_r[UnitEnvironment env, CompoundUnit unit] {
  ExpandedName exp;
  VariableName var;
  Local local;
  String vname;
}: ((WILD JIAZZI_to local=local_r[env,unit] AT WILD {
  if (local != null) {
    for (Iterator i = local.unit().lexemes().values().iterator(); i.hasNext(); i = i) {
      var = (VariableName) i.next();
      if (!unit.lexemes().containsKey(var.text())) continue;
      if (local.lexemeLinkage().containsKey(var)) continue;
      local.link(var, new ExpandedName("", unit.lexemes().get(var.text()), ""));
    }
  }
}) | (exp=expanded_r[env,unit] JIAZZI_to {
} ((local=local_r[env,unit] AT vname=id_r[] {
  if (local != null) {
    if (!local.unit().lexemes().containsKey(vname)) {
      this.reportError("cannot find " + vname + " in local " + local.name());
       
    } else {
      var = local.unit().lexemes().get(vname);
      if (!local.lexemeLinkage().containsKey(var))
	local.link(var, exp);
    }
  }
}) | (WILD AT vname=id_r[] {
  for (Iterator i = unit.locals().values().iterator(); i.hasNext(); i = i) {
    local = (Local) i.next();
    if (!local.unit().lexemes().containsKey(vname)) continue;
    var = local.unit().lexemes().get(vname);
    if (!local.lexemeLinkage().containsKey(var))
      local.link(var, exp);
  }
}))));

*/

link_pkg_r[UnitEnvironment env, CompoundUnit unit] {
  LocPackage from=null;
  LocPackage to=null;
  LocRoot local=null;
  PackageName pkgname=null;
}: ((WILD JIAZZI_to local=local_r[env,unit] AT WILD {
  if (local != null) {
    for (Iterator i = local.packages().values().iterator(); i.hasNext(); /* i = i */) {
      to = (LocPackage) i.next(); 
      if (!to.isSink()) continue;
      if (local.pkgLinkage().containsKey(to)) continue;
      if (!unit.self().packages().containsKey(to.name())) continue;
      from = (LocPackage) unit.self().packages().get(to.name());
      if (!from.isSource()) continue;
      local.link(to, from);
    }
  }
}) | (((ID AT) => local=local_r[env,unit] AT | { local = unit.self(); }) pkgname=pkgname_r[] {
  from = null;
  if (local != null) {
    if (!local.packages().containsKey(pkgname)) {
      this.reportError("cannot find package " + pkgname + " in local " + local.name());
       
    } else {
      from = (LocPackage) local.packages().get(pkgname);
      if (!from.isSource() && !from.isOpen()) {
	this.reportError(local.packages().get(pkgname) + " in local " + local.name() + " is not source");
	 
      } 
    }
  }
} JIAZZI_to ((WILD AT pkgname=pkgname_r[] {
  for (Iterator i = unit.allLocals(); i.hasNext(); /* i = i */) {
    local = (LocRoot) i.next();
    if (!local.packages().containsKey(pkgname)) continue;
    to = (LocPackage) local.packages().get(pkgname);
    if (!to.isSink()) continue;
    if (local.pkgLinkage().containsKey(to)) continue;
    if (from != null) {
      if (from.isOpen())
	local.link(to, from.out());
      else local.link(to, from);
    }
  }
}) | (((ID AT) => local=local_r[env,unit] AT | { local = unit.self(); }) pkgname=pkgname_r[] {

  if (!local.packages().containsKey(pkgname)) {
    this.reportError("package " + pkgname + " not in " + local);
     
  } else {
    to = (LocPackage) local.packages().get(pkgname);

    if (to.isOpen()) {
      if (from == null) {
	this.reportError("cannot connect wild to open package " + pkgname);
	 
      } else if (!from.isOpen()) {
	this.reportError("cannot connect " + from + " to open package " + pkgname);
	 
      } else {
	local.link(to.in(), from.out());
      }
    } else if (!to.isSink()) {
      this.reportError(to + " is not a sink");
    } else if (!local.pkgLinkage().containsKey(to)) 
      if (from != null) {
	if (from.isOpen())
	  local.link(to, from.out());
	else
	  local.link(to, from);
      }
  }
}))));


